import datetime
import errno
import itertools
import math
import os
import re
import shutil
import sqlite3
import time
from argparse import _ArgumentGroup, Namespace
from binascii import hexlify
import random
import string
from pathlib import Path

from clint.textui import progress
from functools import reduce
import numpy as np

from knowsmore.cmdbase import CmdBase
from knowsmore.password import Password
from knowsmore.util.color import Color
from knowsmore.util.database import Database
from knowsmore.util.knowsmoredb import KnowsMoreDB
from knowsmore.util.logger import Logger
from knowsmore.util.tools import Tools

LEETS1 = {"a": "aA@49\u00e1\u00c1\u00e0\u00c0\u00c2\u00c3\u00c4\u00c5\u00e2\u00e3\u00e4\u00e5", "b": "bB8", "c": "cC\u00e7\u00c7", "d": "dD", "e": "eE32\u00c8\u00c9\u00ca\u00cb\u00e8\u00e9\u00ea\u00eb", "f": "fF", "g": "gG96", "h": "hH#", "i": "iI!1\u00cc\u00cd\u00ce\u00cf\u00ec\u00ed\u00ee\u00ef", "j": "jJ", "k": "kK", "l": "lL!1", "m": "mM", "n": "nN\u00f1\u00d1", "o": "oO04\u00d2\u00d3\u00d4\u00d5\u00d6\u00f2\u00f3\u00f4\u00f5\u00f6", "p": "pP", "q": "qQ", "r": "rR", "s": "sS5$", "t": "tT7+", "u": "uU\u00d9\u00da\u00db\u00dc\u00f9\u00fa\u00fb\u00fc", "v": "vV", "w": "wW", "x": "xX", "y": "yY\u00dd\u00fd", "z": "zZ2", "0": "0", "1": "1", "2": "2", "3": "3", "4": "4", "5": "5", "6": "6", "7": "7", "8": "8", "9": "9"}
LEETS2 = {"a":"a","b":"b","c":"c","d":"d","e":"e","f":"f","g":"g","h":"h","i":"i","j":"j","k":"k","l":"l","m":"m","n":"n","o":"o","p":"p","q":"q","r":"r","s":"s","t":"t","u":"u","v":"v","w":"w","x":"x","y":"y","z":"z","A":"A","B":"B","C":"C","D":"D","E":"E","F":"F","G":"G","H":"H","I":"I","J":"J","K":"K","L":"L","M":"M","N":"N","O":"O","P":"P","Q":"Q","R":"R","S":"S","T":"T","U":"U","V":"V","W":"W","X":"X","Y":"Y","Z":"Z","0":"0","1":"1","2":"2","3":"3","4":"4","5":"5","6":"6","7":"7","8":"8","9":"9"}
SPECIAL = ["@", "#", "!", ".", "-", ",", "$", "%", "*", "+", "/", "\\", "<", ">", "=", "&"]
COMMON = ["123", "1234", "12345", "123456"]

class WordList(CmdBase):
    db = None
    name = ''
    min_size = 1
    max_size = 32
    padding = False
    no_leets = False
    unique_chars = []
    unique_ch_b = 0
    char_space = LEETS1
    filename = None
    batch = False
    check_database = False
    append_file = False

    def __init__(self):
        super().__init__('word-list', 'Generates a wordlist based on one word (generally, company name)')

    def add_flags(self, flags: _ArgumentGroup):
        flags.add_argument('--batch',
                           action='store_true',
                           default=False,
                           dest=f'batch',
                           help=Color.s(
                               'Never ask for user input, use the default behavior'))

        flags.add_argument('--append',
                           action='store_true',
                           default=False,
                           dest=f'append_file',
                           help=Color.s(
                               'Append to the output file'))

    def add_commands(self, cmds: _ArgumentGroup):
        cmds.add_argument('--name', action='store', dest='name', help='Name')
        cmds.add_argument('-o', '--out-file', action='store', default='', dest=f'out_file',
                          help=Color.s('Write generate wordlist to a file'))
        cmds.add_argument('-min', '--min-lenght', default=1, type=int, dest='min_lenght',
                          help='Minumin word lenght')
        cmds.add_argument('-max', '--max-lenght', default=32, type=int, dest='max_lenght',
                          help='Maximum word lenght.')
        cmds.add_argument('-p', '--padding', action='store_true',  dest='padding', default=False,
                          help='Add padding to fill string to match minimun leght')
        cmds.add_argument('-nol', '--no-leets', action='store_true', dest='no_leets', default=False,
                          help='No Leets')

    def load_from_arguments(self, args: Namespace) -> bool:

        if args.out_file is None or args.out_file.strip() == '' or \
                args.name is None or args.name.strip() == '':
            Tools.mandatory()

        if not os.path.isdir(Path(args.out_file).parent):
            Logger.pl('{!} {R}error: Output filename is invalid {O}%s{R} {W}\r\n' % (
                args.out_file))
            Tools.exit_gracefully(1)

        self.name = args.name
        self.min_size = int(args.min_lenght)
        self.max_size = int(args.max_lenght)
        self.padding = args.padding
        self.no_leets = args.no_leets
        self.batch = args.batch
        self.append_file = args.append_file

        if self.min_size < 1:
            self.min_size = 1

        if self.no_leets:
            self.char_space = LEETS2

        self.unique_chars = set([v for l1 in [list(value) for value in self.char_space.values()] for v in l1])
        self.unique_ch_b = int(np.sum([len(v.encode("UTF-8")) for v in self.unique_chars]))

        try:
            with open(args.out_file, 'a', encoding="UTF-8") as f:
                pass
        except IOError as x:
            if x.errno == errno.EACCES:
                Logger.pl('{!} {R}error: could not open NTLM hashes file {O}permission denied{R}{W}\r\n')
                Tools.exit_gracefully(1)
            elif x.errno == errno.EISDIR:
                Logger.pl('{!} {R}error: could not open NTLM hashes file {O}it is an directory{R}{W}\r\n')
                Tools.exit_gracefully(1)
            else:
                Logger.pl('{!} {R}error: could not open NTLM hashes file {W}\r\n')
                Tools.exit_gracefully(1)

        self.filename = args.out_file

        # Database not yet needed
        #self.db = self.open_db(args)

        return True

    def run(self):

        estimated_size = self.calculate()
        Logger.pl(
            '{+} {W}KnowsMore will generate +- the following amount of data: {O}%s{W}.' % Tools.sizeof_fmt(
                estimated_size, start_unit="K")
        )
        stat = shutil.disk_usage(Path(self.filename).parent)
        if stat.free/1024 < estimated_size + (estimated_size * 0.2):
            Logger.pl('{!} {R}error: not enough disk space.{W}\r\n')
            Tools.exit_gracefully(1)

        # 3GB
        if estimated_size > 1073741824 and not self.batch:

            Logger.p(
                '{!} {W}Do you want continue? (Y/n): {W}')
            c = input()
            if c.lower() != 'y' and c.lower() != '':
                exit(0)
            print(' ')

        Color.pl('{?} {W}{D}Generating the wordlist wit +- {G}%s{W}{D}, please wait...{W}' % Tools.sizeof_fmt(estimated_size, start_unit="K"))
        count = 0
        lines = 0
        last = 0
        try:
            with open(self.filename, 'w' if not self.append_file else 'a', encoding="UTF-8") as f:
                with progress.Bar(label=" Generating ", expected_size=(estimated_size * 1024.0) + 1, every=1024) as bar:
                    try:
                        for w in self.generate(self.name, 0):
                            txt = f'{w}\n'
                            count += len(txt)
                            lines += 1

                            f.write(f'{w}\n')

                            if count & (1024 * 100) == 0:
                                if count > last:
                                    if count > bar.expected_size:
                                        bar.expected_size = count
                                    bar.show(count)
                                    last = count

                    except KeyboardInterrupt as e:
                        raise e
                    finally:
                        bar.hide = True
                        Tools.clear_line()

        except KeyboardInterrupt as e:
            raise e
        except IOError as x:
            if x.errno == errno.EACCES:
                Logger.pl('{!} {R}error: could not open NTLM hashes file {O}permission denied{R}{W}\r\n')
                Tools.exit_gracefully(1)
            elif x.errno == errno.EISDIR:
                Logger.pl('{!} {R}error: could not open NTLM hashes file {O}it is an directory{R}{W}\r\n')
                Tools.exit_gracefully(1)
            else:
                Logger.pl('{!} {R}error: could not open NTLM hashes file {W}\r\n')
                Tools.exit_gracefully(1)
        finally:
            file_stats = os.stat(self.filename).st_size
            Logger.pl('{+} {W}Generate {O}%d{W} lines ({O}%s{W}) to {G}%s{W}' % (lines, Tools.sizeof_fmt(file_stats), self.filename))

    def calculate(self) -> int:
        s = len(self.name)
        leet_lines = np.prod([len([chars for chars in self.char_space.get(c)]) for c in self.name])
        padding_space = 0
        if len(self.name) < self.min_size and self.padding:
            s1 = self.min_size - len(self.name)
            if s1 > 0:
                pl = math.pow(len(self.unique_chars), s1)
                padding_space = (
                    ((pl * leet_lines) * (s1 + s + 1))
                    + (math.pow(self.unique_ch_b - len(self.unique_chars), s1) * (self.unique_ch_b - len(self.unique_chars)) * (s1 + 1))
                ) * 2
        common_space = np.sum([len(line) + 1 for line in self.add_common(self.name)])

        r = int(
            float(s + 1)/1024.0 +
            float(leet_lines * (s + 1))/1024.0 +
            float(leet_lines * common_space)/1024.0 +
            float(padding_space)/1024.0
        )
        if r < 0:
            r = 0
        return r

    def generate(self, word, index) -> list:
        c = word[index:index + 1]
        if c in self.char_space:
            for i, s in enumerate(self.char_space.get(c)):
                if index == len(word) - 1:
                    p = "%s%s" % (word[0:index], s)
                    if len(p) < self.min_size and self.padding:
                        yield from self.add_padding(p)
                    else:
                        yield p
                    yield from self.add_common(p)
                else:
                    p = "%s%s%s" % (word[0:index], s, word[index + 1:])
                    yield from self.generate(p, index + 1)

    def add_padding(self, word):
        s1 = self.min_size - len(word)
        if s1 > 0:
            for c in list(self.permutation(self.unique_chars, s1)):
                yield "%s%s" % (word, c)
                yield "%s%s" % (c, word)

    def add_common(self, word):

        year = datetime.datetime.now().year
        y2 = int(str(year)[2:4])

        for c in COMMON:
            if self.min_size <= len(word) + len(c) <= self.max_size:
                yield "%s%s" % (word, c)
                yield "%s%s" % (c, word)

        if self.min_size <= len(word) + 1 <= self.max_size:
            for s in SPECIAL:
                yield "%s%s" % (word, s)
                yield "%s%s" % (s, word)

                for c in COMMON:
                    if self.min_size <= len(word) + 1 + len(c) <= self.max_size:
                        yield "%s%s%s" % (word, s, c)
                        yield "%s%s%s" % (c, s, word)
                        yield "%s%s%s" % (word, c, s)
                        yield "%s%s%s" % (s, c, word)

        if self.min_size <= len(word) + 3 <= self.max_size:
            for n in range(0, y2 + 15):
                yield "%s%s" % (word, n)
                yield "%s%s" % (n, word)
                for s in SPECIAL:
                    yield "%s%s%s" % (word, s, n)
                    yield "%s%s%s" % (word, n, s)
                    yield "%s%s%s" % (n, s, word)
                    yield "%s%s%s" % (s, n, word)
                    yield "%s%s%02d" % (word, s, n)
                    yield "%s%02d%s" % (word, n, s)
                    yield "%02d%s%s" % (n, s, word)
                    yield "%s%02d%s" % (s, n, word)

        if self.min_size <= len(word) + 5 <= self.max_size:
            for n in range(year - 15, year + 15):
                yield "%s%s" % (word, n)
                yield "%s%s" % (n, word)
                for s in SPECIAL:
                    yield "%s%s%s" % (word, s, n)
                    yield "%s%s%s" % (word, n, s)
                    yield "%s%s%s" % (n, s, word)
                    yield "%s%s%s" % (s, n, word)
                    yield "%s%s%04d" % (word, s, n)
                    yield "%s%04d%s" % (word, n, s)
                    yield "%04d%s%s" % (n, s, word)
                    yield "%s%04d%s" % (s, n, word)

    def permutation(self, char_space: list, size: int) -> list:
        if size <= 0:
            return []

        # Create a temp array that will be used by
        # allLexicographicRecur()
        data = [""] * (size + 1)

        # Sort the input string so that we get all output strings in
        # lexicographically sorted order
        source = sorted(char_space)

        # Now print all permutations
        yield from self._perm(source, data, size - 1, 0)

    def _perm(self, char_space: list, data: list, size: int, index: int):
        length = len(char_space)

        # One by one fix all characters at the given index and
        # recur for the subsequent indexes
        for i in range(length):

            # Fix the ith character at index and if this is not
            # the last index then recursively call for higher
            # indexes
            data[index] = char_space[i]

            # If this is the last index then print the string
            # stored in data[]
            if index == size:
                yield ''.join(data)
            else:
                yield from self._perm(char_space, data, size, index + 1)







#!/usr/bin/python3
# -*- coding: UTF-8 -*-
import os
import string, random, sys, re
import unicodedata
from tabulate import _table_formats, tabulate

from knowsmore.util.color import Color


class Tools:

    def __init__(self):
        pass

    @staticmethod
    def random_generator(size=6, chars=string.ascii_uppercase + string.digits):
        return ''.join(random.choice(chars) for x in range(size))

    @staticmethod
    def clear_line():
        sys.stderr.write("\033[K")
        sys.stdout.write("\033[K")  # Clear to the end of line

        try:
            size = os.get_terminal_size(fd=os.STDOUT_FILENO)
        except:
            size = 50

        print((" " * size), end='\r', flush=True)
        print((" " * size), file=sys.stderr, end='\r', flush=True)

    @staticmethod     
    def permited_char(s):
        if s.isalpha():
            return True
        elif bool(re.match("^[A-Za-z0-9:]*$", s)):
            return True
        elif s == ".":
            return True
        else:
            return False

    @staticmethod
    def mandatory():
        Color.pl('{!} {R}error: missing a mandatory option, use -h help{W}\r\n')
        Tools.exit_gracefully(1)

    @staticmethod
    def exit_gracefully(code=0):
        exit(code)

    @staticmethod
    def count_file_lines(filename: str):
        def _count_generator(reader):
            b = reader(1024 * 1024)
            while b:
                yield b
                b = reader(1024 * 1024)

        with open(filename, 'rb') as fp:
            c_generator = _count_generator(fp.raw.read)
            # count each \n
            count = sum(buffer.count(b'\n') for buffer in c_generator)
            return count + 1

    @staticmethod
    def clear_string(text):
        return ''.join(filter(Tools.permited_char, Tools.strip_accents(text))).strip().lower()

    @staticmethod
    def strip_accents(text):
        try:
            text = unicode(text, 'utf-8')
        except NameError:  # unicode is a default on python 3
            pass

        text = unicodedata.normalize('NFD', text) \
            .encode('ascii', 'ignore').decode("utf-8")

        return str(text).strip()

    @staticmethod
    def get_tabulated(data: list) -> str:

        headers = [(h if h[0:2] != '__' else ' ') for h in data[0].keys()]
        data = [item.values() for item in data]

        return tabulate(data, headers, tablefmt='psql')

    @staticmethod
    def sizeof_fmt(num, suffix="B", start_unit=""):
        started = False
        for unit in ["", "K", "M", "G", "T", "P", "E", "Z"]:
            if started or start_unit.upper() == unit:
                started = True
                if abs(num) < 1024.0:
                    return f"{num:3.1f} {unit}{suffix}"
                num /= 1024.0
        return f"{num:.1f} Y{suffix}"



import errno
import os
import re
import sqlite3
import time
from argparse import _ArgumentGroup, Namespace
from enum import Enum
from clint.textui import progress

from knowsmore.cmdbase import CmdBase
from knowsmore.config import Configuration
from knowsmore.password import Password
from knowsmore.util.color import Color
from knowsmore.util.database import Database
from knowsmore.util.logger import Logger
from knowsmore.util.tools import Tools


class PreComputed(CmdBase):
    filename = ''
    db = None

    def __init__(self):
        super().__init__('pre-computed', 'Import password list')

    def add_flags(self, flags: _ArgumentGroup):
        pass

    def add_commands(self, cmds: _ArgumentGroup):
        cmds.add_argument('--import-passwords',
                          action='store',
                          metavar='[password file]',
                          type=str,
                          dest=f'pwd_file',
                          help=Color.s('Text file with a password list. Each line is one password'))

    def load_from_arguments(self, args: Namespace) -> bool:
        if args.pwd_file is None or args.pwd_file.strip() == '':
            Tools.mandatory()

        if not os.path.isfile(args.pwd_file):
            Logger.pl('{!} {R}error: Filename is invalid {O}%s{R} {W}\r\n' % (
                args.ntlmfile))
            Tools.exit_gracefully(1)

        try:
            with open(args.pwd_file, 'r') as f:
                # file opened for writing. write to it here
                pass
        except IOError as x:
            if x.errno == errno.EACCES:
                Logger.pl('{!} {R}error: could not open NTLM hashes file {O}permission denied{R}{W}\r\n')
                Tools.exit_gracefully(1)
            elif x.errno == errno.EISDIR:
                Logger.pl('{!} {R}error: could not open NTLM hashes file {O}it is an directory{R}{W}\r\n')
                Tools.exit_gracefully(1)
            else:
                Logger.pl('{!} {R}error: could not open NTLM hashes file {W}\r\n')
                Tools.exit_gracefully(1)

        self.filename = args.pwd_file
        self.db = self.open_db(args)

        return True

    def run(self):
        count = 0
        ignored = 0

        total = Tools.count_file_lines(self.filename)
        with progress.Bar(label=" Processing ", expected_size=total) as bar:
            try:
                with open(self.filename, 'r', encoding="UTF-8", errors="surrogateescape") as f:
                    line = f.readline()
                    while line:
                        try:
                            count += 1
                            bar.show(count)

                            if line.endswith('\n'):
                                line = line[:-1]
                            if line.endswith('\r'):
                                line = line[:-1]

                            if line == '':
                                continue

                            try:
                                password = Password(
                                    ntlm_hash='',
                                    clear_text=line
                                )

                                self.db.insert_ignore_one('pre_computed',
                                                          ntlm_hash=password.ntlm_hash,
                                                          md5_hash=password.md5_hash,
                                                          sha1_hash=password.sha1_hash,
                                                          sha256_hash=password.sha256_hash,
                                                          sha512_hash=password.sha512_hash,
                                                          password=password.clear_text,
                                                          )
                            except Exception as e:
                                Tools.clear_line()
                                Logger.pl('{!} {R}error: could not parse/import line {O}%d{R} => {G}%s {W}: %s\r\n' % (count, line, str(e)))
                                raise e

                        #read next line
                        finally:
                            try:
                                line = f.readline()
                            except:
                                pass

            except KeyboardInterrupt as e:
                raise e
            finally:
                bar.hide = True
                Tools.clear_line()
                Logger.pl('{+} {C}Loaded {O}%s{W} lines' % count)

    def get_ntds_columns(self):
        self.print_verbose('Getting file column design')
        limit = 50
        count = 0
        user_index = ntlm_hash_index = -1

        with open(self.filename, 'r', encoding="UTF-8", errors="surrogateescape") as f:
            line = f.readline()
            while line:
                count += 1
                if count >= limit:
                    break

                line = line.lower()
                if line.endswith('\n'):
                    line = line[:-1]
                if line.endswith('\r'):
                    line = line[:-1]

                c1 = line.split(':')
                if len(c1) < 3:
                    continue

                for idx, x in enumerate(c1):
                    if user_index == -1:
                        if '$' in x or '\\' in x:
                            user_index = idx

                    if ntlm_hash_index == -1:
                        hash = re.sub("[^a-f0-9]", '', x.lower())
                        if len(hash) == 32 and hash != "aad3b435b51404eeaad3b435b51404ee":
                            ntlm_hash_index = idx

                if user_index != -1 and ntlm_hash_index != -1:
                    break

                if user_index == -1 or ntlm_hash_index == -1:
                    user_index = ntlm_hash_index = -1

                try:
                    line = f.readline()
                except:
                    pass

        if user_index < 0 or ntlm_hash_index < 0:
            Logger.pl('{!} {R}error: import file format not recognized {W}\r\n')
            Tools.exit_gracefully(1)

        return user_index, ntlm_hash_index



import hashlib
import codecs
import unicodedata

from knowsmore.util.color import Color
from knowsmore.util.tools import Tools
from math import log
from Levenshtein import ratio

LEETS = {"a": "aA@49\u00e1\u00c1\u00e0\u00c0\u00c2\u00c3\u00c4\u00c5\u00e2\u00e3\u00e4\u00e5", "b": "bB8", "c": "cC\u00e7\u00c7", "d": "dD", "e": "eE32\u00c8\u00c9\u00ca\u00cb\u00e8\u00e9\u00ea\u00eb", "f": "fF", "g": "gG96", "h": "hH#", "i": "iI!1\u00cc\u00cd\u00ce\u00cf\u00ec\u00ed\u00ee\u00ef", "j": "jJ", "k": "kK", "l": "lL!1", "m": "mM", "n": "nN\u00f1\u00d1", "o": "oO04\u00d2\u00d3\u00d4\u00d5\u00d6\u00f2\u00f3\u00f4\u00f5\u00f6", "p": "pP", "q": "qQ", "r": "rR", "s": "sS5$", "t": "tT7+", "u": "uU\u00d9\u00da\u00db\u00dc\u00f9\u00fa\u00fb\u00fc", "v": "vV", "w": "wW", "x": "xX", "y": "yY\u00dd\u00fd", "z": "zZ2", "0": "0", "1": "1", "2": "2", "3": "3", "4": "4", "5": "5", "6": "6", "7": "7", "8": "8", "9": "9"}

class Password(object):
    weak_bits = 30
    ntlm_hash = ''
    bytes_password = bytes()
    clear_text = ''
    latin_clear_text = ''
    length = lower = upper = digit = special = latin = 0
    md5_hash = sha1_hash = sha256_hash = sha512_hash = ''
    entropy = 0
    similarity = 0
    leets_cache = {}

    def __init__(self, ntlm_hash: str, clear_text: str):

        if ntlm_hash == '' and clear_text is not None and clear_text != '':
            import hashlib, binascii
            hash = hashlib.new('md4', clear_text.encode('utf-16le')).digest()
            ntlm_hash = binascii.hexlify(hash)
            if isinstance(ntlm_hash, bytes):
                ntlm_hash = ntlm_hash.decode("UTF-8")
            ntlm_hash = ntlm_hash.lower()

        self.ntlm_hash = ntlm_hash
        self.clear_text = clear_text
        self.latin_clear_text = clear_text
        try:
            self.bytes_password = clear_text.encode("latin-1")
        except:
            self.bytes_password = unicodedata.normalize('NFD', clear_text) \
                .encode('latin-1', 'ignore')

        if Password.leets_cache is None:
            Password.leets_cache = {}

        if '$HEX[' in self.clear_text:
            self.bytes_password = codecs.decode(self.clear_text[5:-1], 'hex_codec')

        self.analyze()
        self.cal_hashes()

        if '$HEX[' in self.clear_text:
            try:
                self.clear_text = self.bytes_password.decode("UTF-8")
            except UnicodeDecodeError:
                try:
                    self.latin_clear_text = self.bytes_password.decode("Latin-1")
                except:
                    pass

                if self.length <= 50:
                    Color.pl('{?} {W}{D}parsed password ({W}{C}%s{W}{D}) contains Latin character, '
                             'using HEX encoded ({W}{C}%s{W}{D}).{W}' % (self.latin_clear_text, self.clear_text))

    def analyze(self):

        self.length = len(self.latin_clear_text)

        self.lower = self.upper = self.digit = self.special = self.latin = 0

        # Calculate entropy
        # it raises an error if length is less than 1
        if self.length > 1:
            unique = set(self.bytes_password)
            self.entropy = int(round(self.length * log(len(unique), 2), 0))

        for idx, c in enumerate(self.latin_clear_text):

                # check latin
                try:
                    b = c.encode("Latin-1")
                    tst = b.decode("UTF-8") # will raise error if it is a latin data
                except UnicodeEncodeError:
                    self.special += 1
                    continue
                except UnicodeDecodeError:
                    self.latin += 1
                    continue

                # counting lowercase alphabets
                if c.islower():
                    self.lower += 1

                # counting uppercase alphabets
                elif c.isupper():
                    self.upper += 1

                # counting digits
                elif c.isdigit():
                    self.digit += 1

                # counting the mentioned special characters
                else:
                    self.special += 1



    def cal_hashes(self):
        self.md5_hash = hashlib.md5(self.bytes_password).hexdigest().lower()
        self.sha1_hash = hashlib.sha1(self.bytes_password).hexdigest().lower()
        self.sha256_hash = hashlib.sha256(self.bytes_password).hexdigest().lower()
        self.sha512_hash = hashlib.sha512(self.bytes_password).hexdigest().lower()

    #https://github.com/kolypto/py-password-strength/blob/master/password_strength/stats.py
    @property
    def strength(self) -> int:
        """ Get password strength as a number normalized to range {0 .. 100}.
        Normalization is done in the following fashion:
        1. If entropy_bits <= weak_bits   -- linear in range{0 .. 33} (weak)
        2. If entropy_bits <= weak_bits*2 -- almost linear in range{33 .. 66} (medium)
        3. If entropy_bits > weak_bits*3  -- asymptotic towards 100 (strong)
        :param weak_bits: Minimum entropy bits a medium password should have.
        :type weak_bits: int
        :return: Normalized password strength:
            * <33 is WEAK
            * <66 is MEDIUM
            * >66 is STRONG
        :rtype: float
        """
        WEAK_MAX = 0.333333333

        if self.entropy <= self.weak_bits:
            return int(round((WEAK_MAX * self.entropy / self.weak_bits) * float(100), 0))

        HARD_BITS = self.weak_bits*3
        HARD_VAL = 0.950

        # Here, we want a function that:
        # 1. f(x)=0.333 at x=weak_bits
        # 2. f(x)=0.950 at x=weak_bits*3 (great estimation for a perfect password)
        # 3. f(x) is almost linear in range{weak_bits .. weak_bits*2}: doubling the bits should double the strength
        # 4. f(x) has an asymptote of 1.0 (normalization)

        # First, the function:
        #       f(x) = 1 - (1-WEAK_MAX)*2^( -k*x)

        # Now, the equation:
        #       f(HARD_BITS) = HARD_VAL
        #       1 - (1-WEAK_MAX)*2^( -k*HARD_BITS) = HARD_VAL
        #                        2^( -k*HARD_BITS) = (1 - HARD_VAL) / (1-WEAK_MAX)
        #       k = -log2((1 - HARD_VAL) / (1-WEAK_MAX)) / HARD_BITS
        k = -log((1 - HARD_VAL) / (1-WEAK_MAX), 2) / HARD_BITS
        f = lambda x: 1 - (1-WEAK_MAX)*pow(2, -k*x)

        return int(round((f(self.entropy - self.weak_bits) * float(100)), 0))  # with offset

    def get_leets(self, word, index=0) -> list:
        if index == 0:
            word = word.lower()
        c = word[index:index + 1]
        if c in LEETS:
            for i, s in enumerate(LEETS.get(c)):
                if index == len(word) - 1:
                    p = "%s%s" % (word[0:index], s)
                    yield p
                else:
                    p = "%s%s%s" % (word[0:index], s, word[index + 1:])
                    yield from self.get_leets(p, index + 1)

    def calc_ratio(self, name: str, score_cutoff: float = 0.0) -> int:
        if name == 0:
            name = name.lower()

        str_pass = self.bytes_password.decode("Latin-1")

        if score_cutoff == 0.0 and len(str_pass) > 0:
            score_cutoff = len(name) / float(len(str_pass)) - 0.05

        if score_cutoff < 0.05:
            score_cutoff = 0.05

        if score_cutoff > 0.4:
            score_cutoff = 0.4

        # Use a static cache to increase speed
        if name not in Password.leets_cache.keys():
            Password.leets_cache[name] = [l1 for l1 in self.get_leets(name)]

        self.similarity = sorted(
            [round(ratio(l1, str_pass, score_cutoff=score_cutoff) * float(100)) for l1 in Password.leets_cache[name]]
        )[-1]

        return self.similarity

    def __str__(self):

        strength = self.strength
        s = "Weak"
        if strength >= 66:
            s = "Strong"
        elif strength >= 33:
            s = "Medium"


        return f'''NTLM........: {self.ntlm_hash}
Clear Text..: {self.clear_text}
Latin Text..: {self.latin_clear_text}
MD5.........: {self.md5_hash}
SHA1........: {self.sha1_hash}
SHA256......: {self.sha256_hash}
SHA512......: {self.sha512_hash}
Entropy.....: {self.entropy}
Strength....: {strength} -> {s}
Length......: {self.length}
Lower.......: {self.lower}
Upper.......: {self.upper}
Digit.......: {self.digit}
Special.....: {self.special}
Latin.......: {self.latin}
'''



import os
import sqlite3
import time
from argparse import _ArgumentGroup, Namespace

from knowsmore.cmdbase import CmdBase
from knowsmore.util.color import Color
from knowsmore.util.database import Database
from knowsmore.util.knowsmoredb import KnowsMoreDB
from knowsmore.util.logger import Logger


class CreateDb(CmdBase):
    db_name = ''
    force = False
    check_database = False

    def __init__(self):
        super().__init__('create-db', 'Create an empty database')

    def add_flags(self, flags: _ArgumentGroup):
        flags.add_argument('--force',
                           action='store_true',
                           default=False,
                           dest=f'force',
                           help=Color.s('Force overwrite SQLite local database'))

    def add_commands(self, cmds: _ArgumentGroup):
        pass

    def load_from_arguments(self, args: Namespace) -> bool:
        if args.dbfile is None or args.dbfile.strip() == '':
            Logger.pl('{!} {R}error: filename is invalid {O}%s{R} {W}\r\n' % (
                args.db))
            exit(1)

        self.db_name = os.path.abspath(args.dbfile.strip())
        self.force = args.force

        if os.path.isfile(self.db_name) and not self.force:
            try:
                db = KnowsMoreDB(auto_create=False,
                                 db_name=self.db_name)

                if db.has_data():
                    Logger.pl('{!} {R}error: database already has data, use parameter {O}--force{R} if you want to replace all data {W}\r\n')
                    exit(1)
            except sqlite3.OperationalError as e:
                print(e)
                Logger.pl(
                    '{!} {R}error: the database file exists but is not an SQLite or table structure was not created. Use parameter {O}--force{R} if you want to replace all data {W}\r\n')
                exit(1)
            except Exception as e:
                raise e

        return True

    def run(self):
        if os.path.isfile(self.db_name) and self.force:
            Color.pl(
                '{!} {W}Database exists (you have 10 seconds to abort...) to prevent overwriting.')
            time.sleep(10)

            os.remove(self.db_name)

        Database(auto_create=True,
                 db_name=self.db_name)

        Logger.pl('{+} {C}Database created {O}%s{W}' % self.db_name)




import json
from argparse import _ArgumentGroup, Namespace
from binascii import hexlify
import random
import string
from clint.textui import progress


from knowsmore.cmdbase import CmdBase
from knowsmore.password import Password
from knowsmore.util.color import Color
from knowsmore.util.database import Database
from knowsmore.util.knowsmoredb import KnowsMoreDB
from knowsmore.util.logger import Logger
from knowsmore.util.tools import Tools


class Wipe(CmdBase):
    db = None
    pre_computed = False

    def __init__(self):
        super().__init__('wipe', 'Wipe critical data. (Clear text password and hashes)')

    def add_flags(self, flags: _ArgumentGroup):
        flags.add_argument('--pre-computed',
                           action='store_true',
                           default=False,
                           dest=f'pre_computed',
                           help=Color.s(
                               'Wipe just pre computed passwords(default: {G}false{W}).{W}'))

    def add_commands(self, cmds: _ArgumentGroup):
        pass

    def load_from_arguments(self, args: Namespace) -> bool:
        self.db = self.open_db(args)
        self.pre_computed = args.pre_computed

        return True

    def run(self):
        challenge_text = ''.join(
            [random.choice(string.digits + string.ascii_lowercase) for i in range(6)])
        text = ''
        first = True

        Logger.pl('{!} {R}ALERT!{W} this process will destroy all%s {G}clear text passwords{W} and {G}hashes{W}. '
                  'But will keep user data and statistics.{W}' % (' (pre-computed)' if self.pre_computed else ''))

        while text != challenge_text:
            try:
                if not first:
                    Logger.pl('{!} {R}Wrong challenge text!{W}.')

                print('')
                Logger.p('{O}Type the text {G}%s{O} and press ENTER to confirm: {W}' % challenge_text)
                text = input()
                first = False

            except (KeyboardInterrupt, EOFError):
                print('')
                print('')
                Logger.pl('{!} {O}Wipe cancelled! {R}Nothing was destroyed!{W}')
                raise KeyboardInterrupt()

        if text == challenge_text:
            passwords = self.db.select('passwords')
            pre_cnt = self.db.select_count('pre_computed')

            count = pre_cnt
            print('')
            Color.pl('{?} {W}{D}destroying passwords data...{W}' + ' ' * 50)
            self.db.delete('pre_computed')
            if not self.pre_computed:
                with progress.Bar(label="Processing ", expected_size=len(passwords)) as bar:
                    try:
                        for idx, pwd in enumerate(passwords):
                            bar.show(idx)

                            fake_data = f"WIPE_{idx}"
                            self.db.update('passwords',
                                           filter_data={'password_id': pwd['password_id']},
                                           ntlm_hash=fake_data,
                                           md5_hash=fake_data,
                                           sha1_hash=fake_data,
                                           sha256_hash=fake_data,
                                           sha512_hash=fake_data,
                                           password=fake_data if pwd['length'] > 0 else '',
                                           )
                            count = pre_cnt + idx

                    except (KeyboardInterrupt, EOFError):
                        bar.hide = True
                        Tools.clear_line()
                        print('')
                        print('')
                        Logger.pl('{!} {O}Wipe cancelled! {G}%d{O} from {G}%d{O} data was {R}destroyed{O}!{W}' %
                                  (count, len(passwords)))
                        raise KeyboardInterrupt()
                    finally:
                        bar.hide = True
                        Tools.clear_line()

            if self.pre_computed:
                Logger.pl('{+} {O}Only Pre-computed passwords and hashes wiped! {G}%d{R} data was destroyed!{W}' % count)
            else:
                Logger.pl('{+} {O}Passwords and hashes wiped! {G}%d{R} data was destroyed!{W}' % count)










import errno
import os
import re
import sqlite3
import time
from argparse import _ArgumentGroup, Namespace
from enum import Enum
from clint.textui import progress

from knowsmore.cmdbase import CmdBase
from knowsmore.config import Configuration
from knowsmore.password import Password
from knowsmore.util.color import Color
from knowsmore.util.database import Database
from knowsmore.util.logger import Logger
from knowsmore.util.tools import Tools


class NTLMHash(CmdBase):
    class ImportMode(Enum):
        Undefined = 0
        NTDS = 1
        Cracked = 2
        Password = 3
        ExportHashes = 4

    filename = ''
    db = None
    mode = ImportMode.Undefined
    password = None

    def __init__(self):
        super().__init__('ntlm-hash', 'Import NTLM hashes and users')

    def add_flags(self, flags: _ArgumentGroup):
        flags.add_argument('-format',
                           action='store',
                           default='secretsdump',
                           dest=f'file_format',
                           help=Color.s(
                               'Specify NTLM hashes format (default: {G}secretsdump{W}). Available methods: {G}secretsdump{W}'))

    def add_commands(self, cmds: _ArgumentGroup):
        cmds.add_argument('--export-hashes',
                          action='store',
                          metavar='[hashes file]',
                          type=str,
                          dest=f'export_file',
                          help=Color.s('NTLM hashes filename.'))

        cmds.add_argument('--import-ntds',
                          action='store',
                          metavar='[hashes file]',
                          type=str,
                          dest=f'ntlmfile',
                          help=Color.s('NTLM hashes filename.'))

        cmds.add_argument('--import-cracked',
                          action='store',
                          metavar='[cracked file]',
                          type=str,
                          dest=f'crackedfile',
                          help=Color.s('Hashcat cracked hashes filename. (format: {G}hash{R}:{G}password{W})'))

        cmds.add_argument('--add-password',
                          action='store',
                          metavar='[clear text password]',
                          type=str,
                          default='',
                          dest=f'password',
                          help=Color.s('Add clear text password to database'))

    def load_from_arguments(self, args: Namespace) -> bool:

        if args.password != '':
            if args.password.strip() == '':
                Tools.mandatory()

            self.password = Password(
                ntlm_hash='',
                clear_text=args.password
            )

            self.mode = NTLMHash.ImportMode.Password

        elif args.export_file is not None and args.export_file != '':
            try:
                with open(args.export_file, 'a') as f:
                    # file opened for writing. write to it here
                    pass
            except IOError as x:
                if x.errno == errno.EACCES:
                    Logger.pl('{!} {R}error: could not open NTLM hashes file {O}permission denied{R}{W}\r\n')
                    Tools.exit_gracefully(1)
                elif x.errno == errno.EISDIR:
                    Logger.pl('{!} {R}error: could not open NTLM hashes file {O}it is an directory{R}{W}\r\n')
                    Tools.exit_gracefully(1)
                else:
                    Logger.pl('{!} {R}error: could not open NTLM hashes file {W}\r\n')
                    Tools.exit_gracefully(1)

            self.mode = NTLMHash.ImportMode.ExportHashes
            self.filename = args.export_file

        else:
            if (args.ntlmfile is None or args.ntlmfile.strip() == '') and \
                    (args.crackedfile is None or args.crackedfile.strip() == ''):
                Tools.mandatory()

            if args.ntlmfile is not None and args.ntlmfile.strip() != '':
                if not os.path.isfile(args.ntlmfile):
                    Logger.pl('{!} {R}error: NTLM filename is invalid {O}%s{R} {W}\r\n' % (
                        args.ntlmfile))
                    Tools.exit_gracefully(1)

                try:
                    with open(args.ntlmfile, 'r') as f:
                        # file opened for writing. write to it here
                        pass
                except IOError as x:
                    if x.errno == errno.EACCES:
                        Logger.pl('{!} {R}error: could not open NTLM hashes file {O}permission denied{R}{W}\r\n')
                        Tools.exit_gracefully(1)
                    elif x.errno == errno.EISDIR:
                        Logger.pl('{!} {R}error: could not open NTLM hashes file {O}it is an directory{R}{W}\r\n')
                        Tools.exit_gracefully(1)
                    else:
                        Logger.pl('{!} {R}error: could not open NTLM hashes file {W}\r\n')
                        Tools.exit_gracefully(1)

                self.mode = NTLMHash.ImportMode.NTDS
                self.filename = args.ntlmfile

            elif args.crackedfile is not None or args.crackedfile.strip() != '':
                if not os.path.isfile(args.crackedfile):
                    Logger.pl('{!} {R}error: NTLM filename is invalid {O}%s{R} {W}\r\n' % (
                        args.ntlmfile))
                    Tools.exit_gracefully(1)

                try:
                    with open(args.crackedfile, 'r') as f:
                        # file opened for writing. write to it here
                        pass
                except IOError as x:
                    if x.errno == errno.EACCES:
                        Logger.pl('{!} {R}error: could not open NTLM hashes file {O}permission denied{R}{W}\r\n')
                        Tools.exit_gracefully(1)
                    elif x.errno == errno.EISDIR:
                        Logger.pl('{!} {R}error: could not open NTLM hashes file {O}it is an directory{R}{W}\r\n')
                        Tools.exit_gracefully(1)
                    else:
                        Logger.pl('{!} {R}error: could not open NTLM hashes file {W}\r\n')
                        Tools.exit_gracefully(1)

                self.mode = NTLMHash.ImportMode.Cracked
                self.filename = args.crackedfile

            if self.mode == NTLMHash.ImportMode.Undefined:
                Logger.pl('{!} {R}error: Nor {O}--import-ntds{R} or {O}--import-cracked{R} was provided{W}\r\n')
                Tools.exit_gracefully(1)

        self.db = self.open_db(args)

        return True

    def run(self):
        if self.mode == NTLMHash.ImportMode.ExportHashes:
            rows = self.db.select_raw(
                sql='select distinct ntlm_hash from passwords',
                args=[]
            )

            Color.pl('{?} {W}{D}Exporting {O}%d{W} hashes...{W}' % len(rows))

            try:
                with open(self.filename, 'w', encoding="UTF-8") as f:
                    for row in rows:
                        f.write(f'{row["ntlm_hash"]}\n')
                    pass
            except KeyboardInterrupt as e:
                raise e
            except IOError as x:
                if x.errno == errno.EACCES:
                    Logger.pl('{!} {R}error: could not open NTLM hashes file {O}permission denied{R}{W}\r\n')
                    Tools.exit_gracefully(1)
                elif x.errno == errno.EISDIR:
                    Logger.pl('{!} {R}error: could not open NTLM hashes file {O}it is an directory{R}{W}\r\n')
                    Tools.exit_gracefully(1)
                else:
                    Logger.pl('{!} {R}error: could not open NTLM hashes file {W}\r\n')
                    Tools.exit_gracefully(1)
            finally:
                Logger.pl('{+} {O}%d{W}{C} exported to {G}%s{W}' % (len(rows), self.filename))

        elif self.mode == NTLMHash.ImportMode.Password:

            pdata = {}

            if Configuration.company != '':
                pdata['company_similarity'] = self.password.calc_ratio(Configuration.company)

            self.db.insert_password_manually(self.password, **pdata)
            Logger.pl('{+} {C}Password inserted/updated{W}')

            print(' ')
            Color.pl('{?} {W}{D}Password data:{W}')
            print(self.password)

            Color.pl('{?} {W}{D}Looking for user with this password...{W}')

            sql = (
                'select c.credential_id, c.name, c.type, c.object_identifier, c.dn, d.domain_id, d.name as domain_name, d.object_identifier as domain_object_identifier, '
                'd.dn as domain_dn, p.password, p.ntlm_hash, p.md5_hash, p.sha1_hash, p.sha256_hash, p.sha512_hash '
                'from credentials as c '
                'inner join passwords as p '
                'on c.password_id = p.password_id '
                'inner join domains as d '
                'on c.domain_id = d.domain_id '
                ' where p.ntlm_hash like ? '
                ' order by c.name'
            )
            args = [self.password.ntlm_hash]

            rows = self.db.select_raw(
                sql=sql,
                args=args
            )

            if len(rows) == 0:
                Logger.pl('{!} {O}Password/hash inserted but none user found with this password{W}\r\n')
                exit(0)

            print(Tools.get_tabulated(rows))

            Logger.pl('{+} {O}%s{W}{C} register found{W}' % len(rows))

        elif self.mode == NTLMHash.ImportMode.NTDS:
            (user_index, ntlm_hash_index) = self.get_ntds_columns()
            min_col = -1
            if user_index > min_col:
                min_col = user_index + 1
            if ntlm_hash_index > min_col:
                min_col = ntlm_hash_index + 1

            count = 0
            ignored = 0
            pc_count = 0

            total = Tools.count_file_lines(self.filename)

            with progress.Bar(label=" Processing ", expected_size=total) as bar:
                try:
                    with open(self.filename, 'r', encoding="UTF-8", errors="surrogateescape") as f:
                        line = f.readline()
                        while line:
                            count += 1
                            bar.show(count)

                            line = line.lower()
                            if line.endswith('\n'):
                                line = line[:-1]
                            if line.endswith('\r'):
                                line = line[:-1]

                            c1 = line.split(':')
                            if len(c1) < min_col:
                                ignored += 1
                                continue

                            f1 = c1[user_index]
                            hash = c1[ntlm_hash_index]

                            if '\\' in f1:
                                f1s = f1.strip().split('\\')
                                domain = f1s[0].strip()
                                usr = f1s[1].strip()
                            else:
                                domain = 'default'
                                usr = f1.strip()

                            type = 'U'

                            if usr.endswith('$'):
                                usr = usr[:-1]
                                type = 'M'

                            if domain == '' or usr == '' or hash == '':
                                self.print_verbose(f'Line ignored: {line}')

                            # try to locate this host previously imported by BloodHound
                            domain_id = -1
                            if domain == 'default':
                                d = self.db.select_raw(
                                    sql="select c.domain_id from credentials as c "
                                        "where name = ? and domain_id <> 1 and type = ?",
                                    args=[usr, type]
                                )
                                if len(d) == 1:  # Not permit duplicity
                                    domain_id = d[0]['domain_id']

                            if domain_id == -1:
                                domain_id = self.db.insert_or_get_domain(domain)

                            if domain_id == -1:
                                Tools.clear_line()
                                Logger.pl('{!} {R}error: Was not possible to import the domain {O}%s{R}\r\n' % domain)
                                Tools.exit_gracefully(1)

                            self.db.insert_or_update_credential(
                                domain=domain_id,
                                username=usr,
                                ntlm_hash=hash,
                                type=type,
                            )

                            # check if exists at pre computed hashes
                            pre_computed = self.db.select('pre_computed',
                                                          ntlm_hash=hash
                                                          )
                            if len(pre_computed) > 0:

                                pc_count += 1

                                pdata = {}

                                password = Password(
                                    ntlm_hash=hash,
                                    clear_text=pre_computed[0]['password']
                                )

                                if Configuration.company != '':
                                    pdata['company_similarity'] = password.calc_ratio(Configuration.company)

                                self.db.update_password(
                                    password,
                                    **pdata
                                )

                            try:
                                line = f.readline()
                            except:
                                pass

                except KeyboardInterrupt as e:
                    raise e
                finally:
                    bar.hide = True
                    Tools.clear_line()
                    if pc_count > 0:
                        Logger.pl('{+} {W}{D}Found %d hashes at pre computed table{W}' % pc_count)
                    Logger.pl('{+} {C}Loaded {O}%s{W} lines' % count)

        elif self.mode == NTLMHash.ImportMode.Cracked:
            count = 0
            ignored = 0

            if Configuration.company == '':
                Logger.pl(
                    '{!} {W}It is recommended import cracked passwords using the parameter {O}--company{W} because '
                    'the KnowsMore will calculate the score of similarity of the passwords and Company Name.'
                    )
                Logger.p(
                    '{!} {W}Do you want continue without inform company name? (y/N): {W}')
                c = input()
                if c.lower() != 'y':
                    exit(0)
                print(' ')

            total = Tools.count_file_lines(self.filename)
            with progress.Bar(label=" Processing ", expected_size=total) as bar:
                try:
                    with open(self.filename, 'r', encoding="UTF-8", errors="surrogateescape") as f:
                        line = f.readline()
                        while line:
                            try:
                                count += 1
                                bar.show(count)

                                if line.endswith('\n'):
                                    line = line[:-1]
                                if line.endswith('\r'):
                                    line = line[:-1]

                                c1 = line.split(':', maxsplit=1)
                                if len(c1) != 2:
                                    ignored += 1
                                    continue

                                if c1[0] == '':
                                    continue

                                password = Password(
                                    ntlm_hash=c1[0].lower(),
                                    clear_text=c1[1]
                                )

                                #verify if exists
                                pwd = self.db.select('passwords',
                                                     ntlm_hash=password.ntlm_hash
                                                     )

                                if len(pwd) == 0:
                                    # insert just at pre_computed
                                    self.db.insert_ignore_one('pre_computed',
                                                              ntlm_hash=password.ntlm_hash,
                                                              md5_hash=password.md5_hash,
                                                              sha1_hash=password.sha1_hash,
                                                              sha256_hash=password.sha256_hash,
                                                              sha512_hash=password.sha512_hash,
                                                              password=password.clear_text,
                                                              )
                                    continue

                                pdata = {}

                                if Configuration.company != '':
                                    pdata['company_similarity'] = password.calc_ratio(Configuration.company)

                                self.db.update_password(
                                    password,
                                    **pdata
                                )

                            #read next line
                            finally:
                                try:
                                    line = f.readline()
                                except:
                                    pass

                except KeyboardInterrupt as e:
                    raise e
                finally:
                    bar.hide = True
                    Tools.clear_line()
                    Logger.pl('{+} {C}Loaded {O}%s{W} lines' % count)

    def get_ntds_columns(self):
        self.print_verbose('Getting file column design')
        limit = 50
        count = 0
        user_index = ntlm_hash_index = -1

        with open(self.filename, 'r', encoding="UTF-8", errors="surrogateescape") as f:
            line = f.readline()
            while line:
                count += 1
                if count >= limit:
                    break

                line = line.lower()
                if line.endswith('\n'):
                    line = line[:-1]
                if line.endswith('\r'):
                    line = line[:-1]

                c1 = line.split(':')
                if len(c1) < 3:
                    continue

                for idx, x in enumerate(c1):
                    if user_index == -1:
                        if '$' in x or '\\' in x:
                            user_index = idx

                    if ntlm_hash_index == -1:
                        hash = re.sub("[^a-f0-9]", '', x.lower())
                        if len(hash) == 32 and hash != "aad3b435b51404eeaad3b435b51404ee":
                            ntlm_hash_index = idx

                if user_index != -1 and ntlm_hash_index != -1:
                    break

                if user_index == -1 or ntlm_hash_index == -1:
                    user_index = ntlm_hash_index = -1

                try:
                    line = f.readline()
                except:
                    pass

        if user_index < 0 or ntlm_hash_index < 0:
            Logger.pl('{!} {R}error: import file format not recognized {W}\r\n')
            Tools.exit_gracefully(1)

        return user_index, ntlm_hash_index



#!/usr/bin/python3
# -*- coding: UTF-8 -*-
import datetime
import sys, os.path
import sqlite3
import string, base64
import json
import hashlib
from sqlite3 import Connection, OperationalError, IntegrityError, ProgrammingError

from .color import Color
from .database import Database
from ..password import Password


class KnowsMoreDB(Database):
    dbName = ""

    def __init__(self, auto_create=True, db_name=None):

        if db_name is None:
            db_name = "knowsmore.db"

        super().__init__(
            auto_create=auto_create,
            db_name=db_name
        )
        self.fill_data()

    def has_data(self) -> bool:
        return self.select_count('passwords') > 0

    def check_open(self) -> bool:
        return self.select_count('passwords') >= 0

    def insert_group(self, domain: int, object_identifier: str, name: str, dn: str = '', members: str = '',
                     membership: str = '') -> int:

        if domain == -1:
            raise Exception('Invalid domain')

        if object_identifier is None or object_identifier.strip() == '':
            raise Exception('object_identifier cannot be empty')

        #grp = self.select('groups',
        #                  object_identifier=object_identifier
        #                  )

        #if len(grp) == 0:
        #    self.insert_one('groups',
        self.insert_update_one('groups',
                               domain_id=domain,
                               name=name,
                               object_identifier=object_identifier,
                               dn=dn,
                               members=members,
                               membership=membership
                               )

    def update_password(self, password: Password, **kwargs):

        filter_data = {
            'ntlm_hash': password.ntlm_hash,
        }

        pwd = {
            'password': password.clear_text,
            'length': password.length,
            'entropy': password.entropy,
            'strength': password.strength,
            'upper': password.upper,
            'lower': password.lower,
            'digit': password.digit,
            'special': password.special,
            'latin': password.latin,
            'md5_hash': password.md5_hash,
            'sha1_hash': password.sha1_hash,
            'sha256_hash': password.sha256_hash,
            'sha512_hash': password.sha512_hash,
        }

        pwd.update(kwargs)

        self.update('passwords', filter_data, **pwd)

        # Get all credentials (user/computer) using this password
        credentials = self.select_raw(
            sql='select distinct c.credential_id, c.name, c.full_name, c.user_data_similarity from credentials as c '
                'inner join passwords as p on c.password_id = p.password_id '
                'where c.type = "U" '
                'and c.full_name != "" '
                'and p.ntlm_hash = ?',
            args=[password.ntlm_hash]
        )
        for c in credentials:
            names = [
                n.lower() for n in c['full_name'].split(' ')
                if len(n) > 3
            ]
            score = sorted(
                [password.calc_ratio(n) for n in names]
            )[-1]
            if int(c['user_data_similarity']) != int(score):
                self.update('credentials',
                            filter_data={'credential_id': c['credential_id']},
                            user_data_similarity=score
                            )

    def insert_password_manually(self, password: Password, **kwargs):

        self.insert_ignore_one('pre_computed',
                               ntlm_hash=password.ntlm_hash,
                               password=password.clear_text if password.length > 0 else ''
                               )

        self.update_password(password, **kwargs)

        self.insert_update_one('pre_computed',
                               ntlm_hash=password.ntlm_hash,
                               md5_hash=password.md5_hash,
                               sha1_hash=password.sha1_hash,
                               sha256_hash=password.sha256_hash,
                               sha512_hash=password.sha512_hash,
                               password=password.clear_text,
                               )

    def insert_or_update_bloodhound_object(self, label:str, object_id: str, filter_type: str = 'objectid',  **props):
        self.insert_update_one(
            'bloodhound_objects',
            object_id=object_id,
            filter_type=filter_type,
            object_label=label,
            props=json.dumps(props)
        )

    def insert_or_update_bloodhound_edge(self, source: str, target: str, source_label: str, target_label: str,
                                         edge_type: str, edge_props: str, filter_type: str = 'objectid',
                                         props: dict = {}):

        txt_props = json.dumps(props)
        checksum = hashlib.md5(
            f'{source_label}:{target_label}:{edge_type}:{source}:{target}:{txt_props}'.encode("UTF-8")
        ).hexdigest().lower()

        self.insert_update_one(
            'bloodhound_edge',
            edge_id=checksum,
            source_id=source,
            destination_id=target,
            edge_props=edge_props,
            source_label=source_label,
            target_label=target_label,
            edge_type=edge_type,
            source_filter_type=filter_type,
            updated_date=datetime.datetime.utcnow(),
            props=txt_props
        )

    def insert_or_update_credential(self, domain: int, username: str, ntlm_hash: str,
                                    dn: str = '', groups: str = '', object_identifier: str = '',
                                    type: str = 'U'):
        try:

            # Hard-coded empty password
            update_password = True
            if ntlm_hash is None or ntlm_hash.strip == '':
                update_password = False
                ntlm_hash = '31d6cfe0d16ae931b73c59d7e0c089c0'

            passwd = self.select_first('passwords', ntlm_hash=ntlm_hash)
            password_id = -1 if passwd is None else passwd['password_id']

            if password_id == -1:
                self.insert_one('passwords', domain_id=domain, ntlm_hash=ntlm_hash)

                passwd = self.select_first('passwords', ntlm_hash=ntlm_hash)
                password_id = -1 if passwd is None else passwd['password_id']

            if password_id == -1:
                raise Exception('Password not found at database')

            cred = self.select_first('credentials',
                                     domain_id=domain,
                                     name=username,
                                     type=type
                                     )
            cred_id = -1 if cred is None else cred['credential_id']

            if cred_id == -1:
                self.insert_ignore_one('credentials',
                                       domain_id=domain,
                                       name=username,
                                       password_id=password_id,
                                       dn=dn if dn is not None else '',
                                       object_identifier=object_identifier if object_identifier is not None else '',
                                       groups=groups if groups is not None else '',
                                       type=type
                                       )
            else:
                data = {
                    'domain_id': domain,
                    'name': username,
                    'type': type
                }
                if dn is not None:
                    data['dn'] = dn
                if groups is not None:
                    data['groups'] = groups
                if object_identifier is not None:
                    data['object_identifier'] = object_identifier
                if update_password:
                    data['password_id'] = password_id

                self.update('credentials', {'credential_id': cred_id}, **data)

        except Exception as e:
            Color.pl('{!} {R}Error inserting credential data:{O} %s{W}' % str(e))
        pass

    def insert_or_get_domain(self, domain: str, dn: str = '', object_identifier: str = '') -> int:

        if domain is None or domain.strip() == '':
            raise Exception('Domain cannot be empty')

        domain = domain.lower()
        dn = '' if dn is None else dn.lower()
        object_identifier = '' if object_identifier is None else object_identifier.strip()

        f = {
            '__operator': 'or',
            'name': domain.lower()
        }
        if dn is not None and dn != '':
            f['dn'] = dn

        if f is None and object_identifier is not None and object_identifier != '':
            f['object_identifier'] = object_identifier

        domain_id = self.get_domain(**f)

        if domain_id == -1:
            data = {
                'name': domain.lower()
            }

            if dn is not None and dn != '':
                data['dn'] = dn

            if object_identifier is not None and object_identifier != '':
                data['object_identifier'] = object_identifier

            self.insert_ignore_one('domains', **data)

            domain_id = self.get_domain(**f)

        else:
            data = {
                'name': domain.lower()
            }

            if dn is not None and dn != '':
                data['dn'] = dn

            if object_identifier is not None and object_identifier != '':
                data['object_identifier'] = object_identifier

            self.update('domains', {'domain_id': domain_id}, **data)

        return domain_id

    def get_domain(self, **kwargs) -> int:
        dom = self.select_first('domains', **kwargs)
        return dom.get('domain_id', None) if dom is not None else -1

    def fill_data(self):
        if not self.has_data():
            # Create default empty password hash
            self.insert_ignore_one('passwords',
                                   ntlm_hash='31d6cfe0d16ae931b73c59d7e0c089c0',
                                   password='(empty)'
                                   )
